"""Fixture for the Store tests"""
import datetime
from copy import deepcopy
import datetime as dt

import pytest

from housekeeper.store import models


@pytest.fixture(scope="function")
def minimal_bundle_obj(case_id, timestamp) -> models.Bundle:
    """Return a bundle object"""
    return models.Bundle(name=case_id, created_at=timestamp)


@pytest.fixture(scope="function")
def bundle_data_json(bundle_data):
    """Fixture for bundle data in json format"""
    json_data = deepcopy(bundle_data)
    json_data["created_at"] = json_data["created_at"].isoformat()
    return json_data


@pytest.fixture(scope="function")
def second_timestamp(timestamp) -> datetime.datetime:
    """Return a time stamp in date time format"""
    return timestamp + datetime.timedelta(days=10)


@pytest.fixture(scope="function")
def old_timestamp() -> datetime.datetime:
    """Return a time stamp that is older than a year"""
    return datetime.datetime(2018, 1, 1)


@pytest.fixture(scope="function")
def second_bundle_data(bundle_data, second_sample_vcf, second_family_vcf, second_timestamp) -> dict:
    """Return a bundle similar to bundle_data with updated file paths"""
    second_bundle = deepcopy(bundle_data)
    second_bundle["created_at"] = second_timestamp
    second_bundle["files"][0]["path"] = str(second_sample_vcf)
    second_bundle["files"][1]["path"] = str(second_family_vcf)
    return second_bundle


@pytest.fixture(scope="function")
def other_case() -> str:
    """Return the case id that differs from base fixture"""
    return "angrybird"


@pytest.fixture(scope="function")
def bundle_data_old(
    bundle_data, second_sample_vcf, second_family_vcf, old_timestamp, other_case
) -> dict:
    """Return info for a older bundle with different files and case id as bundle data"""
    _bundle = deepcopy(bundle_data)
    _bundle["name"] = other_case
    _bundle["created_at"] = old_timestamp
    _bundle["files"][0]["path"] = str(second_sample_vcf)
    _bundle["files"][1]["path"] = str(second_family_vcf)
    return _bundle


@pytest.fixture(scope="function")
def time_stamp_now() -> dt.datetime:
    """Time stamp now"""
    return dt.datetime.now()



"""Module for code to include files in housekeeper"""
import hashlib
import logging
import os
from pathlib import Path

from housekeeper.exc import VersionIncludedError
from housekeeper.store.models import Version

BLOCKSIZE = 65536
EMPTY_STR = ""
LOG = logging.getLogger(__name__)


def link_file(file_path: Path, new_path: Path, hardlink: bool = True) -> None:
    """Create a link for a file"""
    if hardlink:
        LOG.debug("Creating hardlink")
        os.link(file_path.resolve(), new_path)
    else:
        LOG.debug("Creating softlink")
        new_path.symlink_to(file_path)
    LOG.info("linked file: %s -> %s", file_path, new_path)


def include_version(global_root: str, version_obj: Version, hardlink: bool = True):
    """Include files in existing bundle version.

    Including a file means to link them into a folder in the root directory
    """
    LOG.info("Use global root path %s", global_root)
    global_root_dir = Path(global_root)
    if version_obj.included_at:
        raise VersionIncludedError(f"version included on {version_obj.included_at}")

    # generate root directory
    version_root_dir = global_root_dir / version_obj.relative_root_dir
    version_root_dir.mkdir(parents=True, exist_ok=True)
    LOG.info("created new bundle version dir: %s", version_root_dir)

    for file_obj in version_obj.files:
        # hardlink file to the internal structure
        file_obj_path = Path(file_obj.path)
        new_path = version_root_dir / file_obj_path.name
        link_file(file_path=file_obj_path, new_path=new_path, hardlink=hardlink)
        file_obj.path = str(new_path).replace(f"{global_root_dir}/", EMPTY_STR, 1)


def checksum(path: Path) -> str:
    """Calculcate checksum for a file."""
    hasher = hashlib.sha1()
    with open(path, "rb") as stream:
        buf = stream.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = stream.read(BLOCKSIZE)
    return hasher.hexdigest()


def link_to_relative_path(file_path: Path, root_path: Path, version: Version) -> None:
    """Link the given absolute path to its path when included in the given version and return the relative path."""
    housekeeper_path: Path = Path(root_path, version.relative_root_dir, file_path.name)
    link_file(file_path=file_path, new_path=housekeeper_path, hardlink=True)


def relative_path(version: Version, file: Path) -> Path:
    return Path(version.relative_root_dir, file.name)



"""Tests for the core cli"""
import logging

import housekeeper
from housekeeper.cli.core import base
from housekeeper.cli.init import init as init_command


def test_cli_title(cli_runner):
    """Test that the title is shown in output when running the version option"""
    # GIVEN a cli runner

    # WHEN asking to see the version
    result = cli_runner.invoke(base, ["--version"])

    # THEN it should display the title of the program
    assert housekeeper.__title__ in result.output


def test_cli_version(cli_runner):
    """Test that the version is shown in output when running the version option"""
    # GIVEN a cli runner
    # WHEN asking to see the version
    result = cli_runner.invoke(base, ["--version"])
    # THEN it should display the version of the program
    assert housekeeper.__version__ in result.output


def test_init_config(config_file, cli_runner):
    """Test init housekeeper with a config file"""
    # GIVEN a config file and a cli runner

    # WHEN calling the CLI
    result = cli_runner.invoke(base, ["--config", str(config_file), "init"])

    # THEN it should read the config and import it to the context
    assert result.exit_code == 0
    assert "Success!" in result.output


def test_init_database(db_uri, db_path, cli_runner, project_dir):
    """Test init housekeeper with cli options"""
    # GIVEN a uri, a non existing database, a project dir and a cli runner
    assert not db_path.exists()

    # WHEN calling the CLI
    result = cli_runner.invoke(base, ["--database", db_uri, "--root", project_dir, "init"])

    # THEN the database should have been created
    assert db_path.exists()
    # THEN success should be communicated
    assert "Success!" in result.output


def test_init_database_no_specified_database(cli_runner):
    """Test init housekeeper without specifying a database"""
    # GIVEN a cli runner

    # WHEN running the init command without any database or project dir
    result = cli_runner.invoke(base, ["init"])

    # THEN command should exit with a one as exit status
    assert result.exit_code == 1
    # THEN it should be communicated that there is no database
    assert "Please point to a database" in result.output


def test_init_database_no_specified_root_dir(cli_runner, db_uri):
    """Test init housekeeper without specifying a root dir"""
    # GIVEN a cli runner

    # WHEN running the init command without any database or project dir
    result = cli_runner.invoke(base, ["--database", db_uri, "init"])

    # THEN command should exit with a one as exit status
    assert result.exit_code == 1
    # THEN it should be communicated that there is no root dir
    assert "Please specify a root dir" in result.output


def test_init_existing_database(cli_runner, base_context, db_path, caplog):
    """Test to initialise a housekeeper instance when db already exists"""
    caplog.set_level(logging.INFO)
    # GIVEN the uri to an initialised database, a project dir and a cli runner
    assert db_path.exists()

    # WHEN trying to intitialize existing db
    result = cli_runner.invoke(init_command, [], obj=base_context)

    # THEN a controlled exit should be made
    assert result.exit_code == 1
    # THEN it should communicate that the database exists
    assert "Database already exists" in caplog.text


def test_override_existing_database(cli_runner, db_path, base_context, caplog):
    """Test init housekeeper database and overwrite existing one"""
    caplog.set_level(logging.INFO)
    # GIVEN the uri to an initialised database, a project dir and a cli runner
    assert db_path.exists()

    # WHEN intitializing and overriding existing db
    result = cli_runner.invoke(init_command, ["--reset"], obj=base_context, input="Yes")

    # THEN it should communicate that the database exists
    assert "Delete existing tables?" in result.output
    # THEN it should communicate success
    assert "Success!" in caplog.text


def test_force_override_existing_database(cli_runner, db_path, base_context, caplog):
    """Test init housekeeper database and overwrite existing one"""
    caplog.set_level(logging.INFO)
    # GIVEN the uri to an initialised database, a project dir and a cli runner
    assert db_path.exists()

    # WHEN intitializing and overriding existing db
    cli_runner.invoke(init_command, ["--force"], obj=base_context)

    # THEN it should communicate success
    assert "Success!" in caplog.text



"""Functions for dealing with files"""

import json
import logging
from json.decoder import JSONDecodeError

import click
from marshmallow.exceptions import ValidationError

from housekeeper.store.api import schema as schemas

LOG = logging.getLogger(__name__)


def load_json(json_str: str) -> dict:
    """Load a json string"""
    LOG.info("Loading json information")
    try:
        data = json.loads(json_str)
    except JSONDecodeError as err:
        LOG.warning("Something wrong in json string")
        LOG.error(err)
        raise click.Abort
    LOG.info("Succesfull loading of JSON")
    return data


def validate_input(data: dict, input_type: str):
    """Validate input with the marshmallow schemas"""
    valid_schemas = {
        "bundle": schemas.InputBundleSchema(),
        "file": schemas.InputFileSchema(),
        "version": schemas.InputVersionSchema(),
    }
    schema = valid_schemas.get(input_type)
    if schema is None:
        LOG.warning("Invalid input type %s", input_type)
        raise ValueError()

    LOG.info("Validating bundle schema")

    formatted_data = schema.dump(data)
    try:
        LOG.info("Validate marshmallow schema")
        schema.load(formatted_data)
    except ValidationError as err:
        LOG.warning("Input data does not follow the models")
        LOG.error(err)
        raise click.Abort
    LOG.info("Input looks fine")



"""Based on https://github.com/pypa/sampleproject/blob/master/setup.py."""
import io

# To use a consistent encoding
import os

# Always prefer setuptools over distutils
from setuptools import find_packages, setup

NAME = "housekeeper"
DESCRIPTION = "Housekeeper takes care of files"
AUTHOR = "Robin Andeer"
EMAIL = "mans.magnusson@scilifelab.se"
URL = "https://github.com/Clinical-Genomics/housekeeper"

HERE = os.path.abspath(os.path.dirname(__file__))


def parse_reqs(req_path="./requirements.txt"):
    """Recursively parse requirements from nested pip files."""
    install_requires = []
    with open(req_path, "r") as handle:
        # remove comments and empty lines
        lines = (line.strip() for line in handle if line.strip() and not line.startswith("#"))
        for line in lines:
            # check for nested requirements files
            if line.startswith("-r"):
                # recursively call this function
                install_requires += parse_reqs(req_path=line[3:])
            else:
                # add the line as a new requirement
                install_requires.append(line)
    return install_requires


# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(HERE, "README.md"), encoding="utf-8") as f:
        LONG_DESCRIPTION = "\n" + f.read()
except FileNotFoundError:
    LONG_DESCRIPTION = DESCRIPTION


setup(
    name=NAME,
    # Versions should comply with PEP440. For a discussion on
    # single-sourcing the version across setup.py and the project code,
    # see http://packaging.python.org/en/latest/tutorial.html#version
    version="4.10.14",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    # What does your project relate to? Separate with spaces.
    keywords="housekeeper development",
    author=AUTHOR,
    author_email=EMAIL,
    license="MIT",
    # The project's main homepage
    url=URL,
    packages=find_packages(exclude=("tests*", "docs", "examples")),
    # If there are data files included in your packages that need to be
    # installed, specify them here.
    include_package_data=True,
    zip_safe=False,
    # Although 'package_data' is the preferred approach, in some case you
    # may need to place data files outside of your packages.
    # In this case, 'data_file' will be installed into:
    # '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],
    # Install requirements loaded from ``requirements.txt``
    install_requires=parse_reqs(),
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and
    # allow pip to create the appropriate form of executable for the
    # target platform.
    entry_points={"console_scripts": ["housekeeper = housekeeper.cli:base"]},
    # See: http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are:
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Environment :: Console",
    ],
)



"""Tests for include cli command"""
import logging
from datetime import datetime
from pathlib import Path

from click.testing import CliRunner
from housekeeper.cli.include import include
from housekeeper.constants import ROOT
from housekeeper.store.api.core import Store
from housekeeper.store.models import Bundle, Version


def test_include_files_creates_bundle_dir(populated_context: dict, cli_runner: CliRunner):
    """Test to include the files of a version for a bundle_name

    The bundle should not exist before and after command is run it should have been created
    """
    # GIVEN a context that is populated
    store: Store = populated_context["store"]
    bundle: Bundle = store._get_query(table=Bundle).first()
    bundle_name: str = bundle.name
    # GIVEN that the latest version of the bundle is not included
    assert bundle.versions[0].included_at is None
    # GIVEN that no folder has been created since case is not included
    bundle_path = Path(populated_context[ROOT], bundle.name)
    assert not bundle_path.exists()

    # WHEN running the include files command
    cli_runner.invoke(include, [bundle_name], obj=populated_context)

    # THEN assert that the bundle path was created
    assert bundle_path.exists()


def test_include_files_creates_version_specific_bundle_dir(
    populated_context: dict, cli_runner: CliRunner
):
    """Test to include the files of a version for a bundle_name

    The version bundle should not exist before and after command is run it should have been created
    """
    # GIVEN a context that is populated
    store: Store = populated_context["store"]
    bundle: Bundle = store._get_query(table=Bundle).first()
    version_obj = bundle.versions[0]
    bundle_name: str = bundle.name
    # GIVEN that the latest version of the bundle is not included
    assert version_obj.included_at is None
    # GIVEN that no version specific folder has been created since version is not included
    version_path: Path = Path(
        populated_context[ROOT], bundle.name, str(version_obj.created_at.date())
    )
    assert not version_path.exists()

    # WHEN running the include files command
    cli_runner.invoke(include, [bundle_name], obj=populated_context)

    # THEN assert that the version path was created
    assert version_path.exists()


def test_include_files_adds_version_specific_files(populated_context: dict, cli_runner: CliRunner):
    """Test to include the files of a version for a bundle_name

    The files should have been hard linked after command has been run
    """
    # GIVEN a context that is populated
    store: Store = populated_context["store"]
    bundle: Bundle = store._get_query(table=Bundle).first()
    version_obj = bundle.versions[0]
    bundle_name: str = bundle.name
    # GIVEN that the latest version of the bundle is not included
    assert version_obj.included_at is None
    # GIVEN that no version specific folder has been created since version is not included
    version_path = Path(populated_context[ROOT], bundle.name, str(version_obj.created_at.date()))
    assert not version_path.exists()

    # WHEN running the include files command
    cli_runner.invoke(include, [bundle_name], obj=populated_context)

    files_included = list(version_path.iterdir())
    assert files_included
    assert len(files_included) == len(version_obj.files)


def test_include_files_specific_version(populated_context: dict, cli_runner: CliRunner):
    """Test to include the files of a version by specifying the version id

    The folder should have been created
    """
    # GIVEN a context that is populated
    store: Store = populated_context["store"]
    bundle: Bundle = store._get_query(table=Bundle).first()
    version_obj = bundle.versions[0]
    version_id = version_obj.id
    # GIVEN that the latest version of the bundle is not included
    assert version_obj.included_at is None
    # GIVEN that no version specific folder has been created since version is not included
    version_path: Path = Path(
        populated_context[ROOT], bundle.name, str(version_obj.created_at.date())
    )
    assert not version_path.exists()

    # WHEN running the include files command
    cli_runner.invoke(include, ["--version-id", version_id], obj=populated_context)

    # THEN assert that the folder was created
    assert version_path.exists()

    files_included = list(version_path.iterdir())
    assert files_included
    assert len(files_included) == len(version_obj.files)


def test_include_version_no_args(populated_context: dict, cli_runner: CliRunner, caplog):
    """Test to include the files of a version without specifying bundle name or version

    The command should exit since it needs to be specified
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context that is populated

    # WHEN running the include files command without args
    result = cli_runner.invoke(include, [], obj=populated_context)

    # THEN assert that the command exists with zero code 1
    assert result.exit_code == 1
    # THEN assert it was communicated that arguments are needed
    assert "use bundle name or version-id" in caplog.text


def test_include_non_existing_version(populated_context: dict, cli_runner: CliRunner, caplog):
    """Test to include the files of a version that does not exist

    The command should exit since a valid version is needed
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context that is populated
    store: Store = populated_context["store"]
    # GIVEN a version that does not exists
    version_id = 10
    assert not store.get_version_by_id(version_id=version_id)

    # WHEN running the include files specifying the non existing version
    result = cli_runner.invoke(include, ["--version-id", version_id], obj=populated_context)

    # THEN assert that the command exists with zero code 1
    assert result.exit_code == 1
    # THEN assert it was communicated that version did not exist
    assert "version not found" in caplog.text


def test_include_non_existing_bundle(
    base_context: dict, cli_runner: CliRunner, timestamp: datetime, caplog
):
    """Test to include the files in the latest version of a bundle that does not exist

    The command should exit since a valid bundle is needed
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN context and a bundle_name that does not exist
    store = base_context["store"]
    bundle_name = "hello"
    assert not store.get_bundle_by_name(bundle_name=bundle_name)

    # WHEN running the include files specifying bundle without versions
    result = cli_runner.invoke(include, [bundle_name], obj=base_context)

    # THEN assert that the command exists with zero code 1
    assert result.exit_code == 1
    # THEN assert it was communicated that the bundle did not exist
    assert f"bundle {bundle_name} not found" in caplog.text


def test_include_bundle_without_version(
    base_context: dict, cli_runner: CliRunner, timestamp: datetime, caplog
):
    """Test to include the files in the latest version of a bundle that does not have versions

    The command should exit since a valid bundle with versions are needed
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN context with a bundle without versions
    store: Store = base_context["store"]
    bundle_name = "hello"

    new_bundle = store.new_bundle(name=bundle_name, created_at=timestamp)
    store.session.add(new_bundle)
    store.session.commit()

    bundle = store.get_bundle_by_name(bundle_name=bundle_name)
    assert not bundle.versions

    # WHEN running the include files specifying bundle without versions
    result = cli_runner.invoke(include, [bundle_name], obj=base_context)

    # THEN assert that the command exists with zero code 1
    assert result.exit_code == 1
    # THEN assert it was communicated that version did not exist
    assert f"Could not find any versions for bundle {bundle_name}" in caplog.text


def test_include_version_already_included(populated_context: dict, cli_runner: CliRunner, caplog):
    """Test to include the files of a version that is already included

    The command should exit since the version can not be included again
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context that is populated and a version that is already included
    store: Store = populated_context["store"]
    version_obj = store._get_query(table=Version).first()
    version_id = version_obj.id
    cli_runner.invoke(include, ["--version-id", version_id], obj=populated_context)

    # WHEN trying to include the version again
    result = cli_runner.invoke(include, ["--version-id", version_id], obj=populated_context)

    # THEN assert that the command exists with zero code 1
    assert result.exit_code == 1
    # THEN assert it was communicated that version did not exist
    assert "version included on" in caplog.text



"""Initialise HK db from CLI"""
import logging

import click

from housekeeper.store.database import create_all_tables, drop_all_tables, get_tables

LOG = logging.getLogger(__name__)


@click.command()
@click.option("--reset", is_flag=True, help="reset database before setting up tables")
@click.option("--force", is_flag=True, help="bypass manual confirmations")
@click.pass_context
def init(context, reset, force):
    """Setup the database."""
    existing_tables: list[str] = get_tables()

    if force or reset:
        if existing_tables and not force:
            message = f"Delete existing tables? [{', '.join(existing_tables)}]"
            click.confirm(click.style(message, fg="yellow"), abort=True)
        drop_all_tables()
    elif existing_tables:
        LOG.error("Database already exists, use '--reset'")
        context.abort()

    create_all_tables()
    LOG.info(f"Success! New tables: {', '.join(get_tables())}")



from sqlalchemy.orm import Query

from housekeeper.store.api.core import Store
from housekeeper.store.models import Version
from housekeeper.store.filters.version_filters import filter_version_by_id


def test_get_version_query_returns_query_object(populated_store: Store):
    result = populated_store._get_query(table=Version)
    assert isinstance(result, Query)


def test_filter_version_by_id_returns_the_correct_version(populated_store: Store):
    """Test getting version by version ID."""

    # GIVEN a store with a version
    version: Version = populated_store._get_query(table=Version).first()
    assert version

    version_id: int = version.id

    # WHEN retrieving the version by ID
    filtered_version: Version = filter_version_by_id(
        versions=populated_store._get_query(table=Version),
        version_id=version_id,
    ).first()

    # THEN a version should be returned
    assert isinstance(filtered_version, Version)

    # THEN the id should match
    assert filtered_version.id == version_id


def test_filter_version_by_id_with_invalid_id_returns_empty_query(
    populated_store: Store,
):
    """Test that an empty result is returned when using an invalid version ID."""

    # GIVEN a store with versions
    versions_query = populated_store._get_query(table=Version)
    assert versions_query.count() > 0

    # GIVEN an invalid id that doesn't exist in the store
    non_existing_version_id = versions_query.order_by(Version.id.desc()).first().id + 1

    # WHEN filtering the versions by the invalid ID
    filtered_versions_query = filter_version_by_id(
        versions=versions_query, version_id=non_existing_version_id
    )

    # THEN the filtered query should be empty
    assert filtered_versions_query.count() == 0


def test_filter_version_by_id_with_empty_query_returns_empty_query(
    populated_store: Store,
):
    """Test that the function returns an empty result when given an empty query."""

    # GIVEN an empty query
    empty_query = populated_store._get_query(table=Version).filter(False)
    assert empty_query.count() == 0

    # WHEN filtering the empty query by an ID
    filtered_query = filter_version_by_id(versions=empty_query, version_id=1)

    # THEN the filtered query should still be empty
    assert filtered_query.count() == 0



"""Tests for adding via CLI"""
from housekeeper.cli.get import bundle_cmd
from housekeeper.store.api.core import Store
from housekeeper.store.models import Bundle


def test_get_existing_bundle_name(populated_context, cli_runner, helpers):
    """Test to fetch an existing bundle based on name"""

    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]

    # GIVEN a existing bundle
    bundle: Bundle = store._get_query(table=Bundle).first()
    assert bundle

    bundle_name = bundle.name

    # WHEN trying to fetch the bundle based on bundle name
    output = helpers.get_stdout(
        cli_runner.invoke(bundle_cmd, [bundle_name], obj=populated_context).output
    )

    # THEN assert that the bundle was written to terminal
    assert bundle_name in output


def test_get_existing_bundle_verbose(populated_context, cli_runner, helpers):
    """Test to fetch an existing bundle based on name with verbose information"""

    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]

    # GIVEN a existing bundle
    bundle: Bundle = store._get_query(table=Bundle).first()
    assert bundle
    bundle_name = bundle.name

    # WHEN trying to fetch the bundle based on bundle name
    output = helpers.get_stdout(
        cli_runner.invoke(bundle_cmd, [bundle_name, "-v"], obj=populated_context).output
    )
    # THEN assert that the files are printed
    assert "Files table" in output


def test_get_non_existing_bundle_name(base_context, cli_runner, helpers, case_id):
    """Test to fetch a non existing bundle based on name when store is empty"""

    # GIVEN a context with a empty store and a cli runner
    store = base_context["store"]

    # GIVEN that there are no bundles
    assert helpers.count_iterable(store.bundles()) == 0

    # WHEN trying to fetch the bundle based on bundle name
    output = helpers.get_stdout(cli_runner.invoke(bundle_cmd, [case_id], obj=base_context).output)

    # THEN assert that no bundle was written to terminal
    assert case_id not in output


def test_get_non_existing_bundle_populated_store(
    populated_context, cli_runner, other_case_id, helpers
):
    """Test to fetch a non existing bundle based on name when bundles exists"""
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]

    # GIVEN a non empty store
    assert helpers.count_iterable(store.bundles()) > 0

    # GIVEN a bundle name that does not exist in database
    assert store.get_bundle_by_name(bundle_name=other_case_id) is None

    # WHEN trying to fetch the non existing bundle
    output = helpers.get_stdout(
        cli_runner.invoke(bundle_cmd, [other_case_id], obj=populated_context).output
    )

    # THEN assert that no bundle was written to terminal
    assert other_case_id not in output


def test_get_existing_bundle_id(populated_context, cli_runner, helpers):
    """Test to fetch an existing bundle based on bundle id"""
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]

    # GIVEN a existing bundle
    bundle = store._get_query(table=Bundle).first()
    bundle_id = bundle.id

    # WHEN trying to fetch a bundle based on bundle id
    json_bundles = helpers.get_json(
        cli_runner.invoke(bundle_cmd, ["-i", bundle_id, "--json"], obj=populated_context).output
    )

    # THEN assert that the bundle was printed to screen
    assert isinstance(json_bundles, list)
    assert len(json_bundles) == 1


def test_get_bundle_json(populated_context, cli_runner, helpers):
    """Test to fetch a bundle in json format"""

    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]

    # GIVEN a existing bundle
    bundle = store._get_query(table=Bundle).first()
    bundle_id = bundle.id

    # WHEN fetching the bundle in json format
    json_bundles = helpers.get_json(
        cli_runner.invoke(bundle_cmd, ["-i", bundle_id, "--json"], obj=populated_context).output
    )

    # THEN assert that the output is a list of bundles
    assert isinstance(json_bundles, list)

    # THEN assert that the bundles are dictionaries
    assert isinstance(json_bundles[0], dict)


def test_get_bundles_multiple_bundles(populated_context, cli_runner, helpers, other_bundle):
    """Test to get all bundles when there are more than one bundle"""

    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]
    helpers.add_bundle(store, other_bundle)

    # GIVEN a store with more than one bundles
    nr_bundles = helpers.count_iterable(store.bundles())
    assert nr_bundles > 1

    # WHEN fetching all bundles by not specifying any bundle
    json_bundles = helpers.get_json(
        cli_runner.invoke(bundle_cmd, ["--json"], obj=populated_context).output
    )

    # THEN assert that all bundles where fetched
    assert len(json_bundles) == nr_bundles


def test_get_bundles_no_bundle(base_context, cli_runner, helpers):
    """Test to get all bundles when there are no bundles"""

    # GIVEN a context with a populated store and a cli runner
    store = base_context["store"]

    # GIVEN a store without bundles
    assert helpers.count_iterable(store.bundles()) == 0

    # WHEN fetching all bundles by not specifying any bundle
    json_bundles = helpers.get_json(
        cli_runner.invoke(bundle_cmd, ["--json"], obj=base_context).output
    )

    # THEN assert that we still get a list
    assert isinstance(json_bundles, list)

    # THEN assert that no bundles where fetched
    assert len(json_bundles) == 0



"""Tests for finding tags in store."""
from datetime import timedelta
from pathlib import Path

import pytest

from housekeeper.store import Store
from housekeeper.store.models import Archive, File, Tag


def test_tag_with_tag_name(populated_store: Store, sample_tag_name: str):
    """Test fetching a tag from database given a tag name."""
    # GIVEN a populated store and a tag name

    # WHEN retrieving a tag from Store
    test_tag = populated_store.get_tag(tag_name=sample_tag_name)

    # THEN a tag should be returned
    assert isinstance(test_tag, Tag)

    # THEN the tag should have the tag name gotten as parameter
    assert test_tag.name == sample_tag_name


def test_tag_without_tag_name(populated_store: Store):
    """Test fetching a tag from database without specifying a tag name returns None."""
    # GIVEN a populated store

    # WHEN retrieving a tag from Store
    test_tag = populated_store.get_tag()

    # THEN a tag should be returned
    assert test_tag is None


def test_get_files_before(populated_store, bundle_data_old, time_stamp_now):
    """
    Test return all files when two bundles are added and all files are older.
    """
    # GIVEN a populated store
    store: Store = populated_store
    starting_nr_of_files: int = len(store.get_files_before(before_date=time_stamp_now))

    # GIVEN two new files
    bundle_old_obj, _ = store.add_bundle(data=bundle_data_old)
    store.session.add(bundle_old_obj)
    store.session.commit()

    # WHEN fetching all files in the database
    files = store.get_files_before(before_date=time_stamp_now)

    # THEN two more files should be returned
    assert len(files) == starting_nr_of_files + 2


def test_get_past_files(populated_store, bundle_data_old, timestamp, old_timestamp):
    """
    test fetch files where not all files are older than before date.
    """
    store: Store = populated_store
    # GIVEN a store with two bundles and two files in each bundle
    bundle_old_obj, _ = store.add_bundle(data=bundle_data_old)
    store.session.add(bundle_old_obj)
    store.session.commit()

    # WHEN fetching all files before the oldest date
    date = old_timestamp + timedelta(days=10)
    assert old_timestamp < date < timestamp
    files: list[File] = store.get_files_before(before_date=date)

    # THEN a list of Files is returned
    assert isinstance(files[0], File)

    # THEN assert only files from the old bundle was found
    assert len(files) == 2


def test_get_no_get_files_before_oldest(populated_store, bundle_data_old, old_timestamp, timestamp):
    """
    Test get files where no files are older than before date.
    """
    store: Store = populated_store
    # GIVEN a store with two bundles and two files in each bundle
    bundle_old_obj, _ = store.add_bundle(data=bundle_data_old)
    store.session.add(bundle_old_obj)
    store.session.commit()

    # WHEN fetching all files before the oldest date
    date = old_timestamp - timedelta(days=10)
    assert date < old_timestamp < timestamp
    files: list[File] = store.get_files_before(before_date=date)

    # THEN assert no files where that old
    assert len(files) == 0


def test_get_archived_files_for_bundle(
    archived_file: Path,
    non_archived_file: Path,
    populated_store: Store,
    sample_id: str,
    spring_tag: str,
):
    """Tests fetching all archive SPRING files in a given bundle."""
    # GIVEN a bundle with two files, where one is archive and one is not

    # WHEN asking for all archived files
    archived_files: list[Path] = [
        Path(file.path)
        for file in populated_store.get_archived_files_for_bundle(
            bundle_name=sample_id, tags=[spring_tag]
        )
    ]

    # THEN only one should be returned
    assert archived_file in archived_files
    assert non_archived_file not in archived_files


def test_get_non_archived_files_for_bundle(
    archived_file: Path,
    non_archived_file: Path,
    populated_store: Store,
    sample_id: str,
    spring_tag: str,
):
    """Tests getting all non-archive SPRING files in a given bundle."""
    # GIVEN a bundle with two files, where one is archive and one is not

    # WHEN asking for all non-archived files
    archived_files: list[Path] = [
        Path(file.path)
        for file in populated_store.get_non_archived_files_for_bundle(
            bundle_name=sample_id, tags=[spring_tag]
        )
    ]

    # THEN only one should be returned
    assert archived_file not in archived_files
    assert non_archived_file in archived_files


def test_get_bundle_name_from_file_path(
    populated_store: Store, sample_id: str, spring_file_1: Path
):
    """Test that the bundle name is fetched correctly from a file path."""
    # GIVEN a store containing a spring file related to sample ACC123456A1

    # WHEN getting the bundle name for the file
    bundle_name: str = populated_store.get_bundle_name_from_file_path(spring_file_1.as_posix())

    # THEN the bundle name should be the sample name
    assert bundle_name == sample_id


@pytest.mark.parametrize("limit", [None, 0])
def test_get_non_archived_files(populated_store: Store, spring_tag: str, limit: int | None):
    """Test that getting all non-archived spring files from the store
    returns all files fulfilling said condition."""
    # GIVEN a populated store containing SPRING and non-SPRING entries
    all_files: list[File] = populated_store.get_files().all()
    assert all_files

    # WHEN retrieving all non archived spring files
    non_archived_spring_files: list[File] = populated_store.get_non_archived_files(
        tag_names=[spring_tag], limit=limit
    )

    # WHEN no limit was set
    if limit is None:
        # THEN entries should be returned
        assert non_archived_spring_files

        # THEN all files with archives and the SPRING tag should be returned
        for file in all_files:
            if file not in non_archived_spring_files:
                assert file.archive or spring_tag not in [tag.name for tag in file.tags]
            else:
                assert not file.archive
                assert spring_tag in [tag.name for tag in file.tags]
                assert file in non_archived_spring_files

    # WHEN a limit is set to not fetch any files
    else:
        # THEN no files should be returned
        assert not non_archived_spring_files


def test_get_ongoing_archiving_tasks(
    archive: Archive, archiving_task_id: int, populated_store: Store
):
    """Tests returning unfinished archiving tasks ids."""
    # GIVEN a populated store with one ongoing archiving task
    archive.archiving_task_id = archiving_task_id
    archive.archived_at = None

    # WHEN getting ongoing archiving tasks
    ongoing_task_ids: set[int] = set(
        archive_entry.archiving_task_id for archive_entry in populated_store.get_ongoing_archivals()
    )

    # THEN the set should include the initial archiving task id
    assert archiving_task_id in ongoing_task_ids


def test_get_ongoing_retrieval_tasks(
    archive: Archive, retrieval_task_id: int, populated_store: Store
):
    """Tests the returning of ongoing retrieval tasks."""
    # GIVEN a populated store with one ongoing retrieval task
    archive.retrieval_task_id = retrieval_task_id
    archive.retrieved_at = None

    # WHEN getting ongoing retrieval tasks
    ongoing_task_ids: set[int] = set(
        archive_entry.retrieval_task_id
        for archive_entry in populated_store.get_ongoing_retrievals()
    )

    # THEN the set should include the initial retrieval task id
    assert retrieval_task_id in ongoing_task_ids


def test_get_archive_entries_by_archiving_id(
    archive: Archive, archiving_task_id: int, populated_store: Store
):
    """Tests that get_archives returns archive entries with the given archiving task id."""

    # GIVEN a populated store with an archive entry

    # WHEN getting entries with the matching archiving task id
    archives: list[Archive] = populated_store.get_archives(archival_task_id=archiving_task_id)

    # THEN the matching entry should have been returned
    assert archive in archives


def test_get_archive_entries_by_retrieval_id(retrieval_task_id: int, populated_store: Store):
    """Tests that get_archives returns archive entries with the given archiving task id."""

    # GIVEN a populated store with an archive entry

    # WHEN getting entries with the matching retrieval task id
    archives: list[Archive] = populated_store.get_archives(retrieval_task_id=retrieval_task_id)

    # THEN an Archive entry should be returned
    assert archives

    # THEN all returned entries should have the specified retrieval task id
    for archive in archives:
        assert archive.retrieval_task_id == retrieval_task_id


def test_archives_not_returned_via_archiving_id(archiving_task_id: int, populated_store: Store):
    """Tests that only archives with matching archiving task ids are returned."""

    # GIVEN a store with multiple archive entries, out of which only one has the provided archiving task id

    # WHEN fetching all archives
    all_archives: list[Archive] = populated_store.get_archives()

    # WHEN fetching all archives with a specified archiving task id
    selected_archives: list[Archive] | None = populated_store.get_archives(
        archival_task_id=archiving_task_id
    )

    # THEN only archives with the matching retrieval task id is returned
    assert len(all_archives) > len(selected_archives)
    for archive in all_archives:
        if archive.archiving_task_id == archiving_task_id:
            assert archive in selected_archives
        else:
            assert archive not in selected_archives


def test_archives_not_returned_via_retrieval_id(retrieval_task_id: int, populated_store: Store):
    """Tests that only archives with matching retrieval task ids are returned."""

    # GIVEN a store with multiple archive entries, out of which only one has the provided retrieval task id

    # WHEN fetching all archives
    all_archives: list[Archive] = populated_store.get_archives()

    # WHEN fetching all archives with a specified retrieval task id
    selected_archives: list[Archive] | None = populated_store.get_archives(
        retrieval_task_id=retrieval_task_id
    )

    # THEN only archives with the matching retrieval task id is returned
    assert len(all_archives) > len(selected_archives)
    for archive in all_archives:
        if archive.retrieval_task_id == retrieval_task_id:
            assert archive in selected_archives
        else:
            assert archive not in selected_archives



"""Module for the models"""

import datetime as dt
from pathlib import Path

from sqlalchemy import Column, ForeignKey, Table, UniqueConstraint, orm, types
from sqlalchemy.orm import backref, declarative_base

Model = declarative_base()

file_tag_link = Table(
    "file_tag_link",
    Model.metadata,
    Column("file_id", types.Integer, ForeignKey("file.id", ondelete="CASCADE"), nullable=False),
    Column("tag_id", types.Integer, ForeignKey("tag.id", ondelete="CASCADE"), nullable=False),
    UniqueConstraint("file_id", "tag_id", name="_file_tag_uc"),
)


class Archive(Model):
    """Information regarding the archiving of a file."""

    __tablename__ = "archive"
    archiving_task_id = Column(types.Integer, nullable=False)
    retrieval_task_id = Column(types.Integer, nullable=True)
    file_id = Column(ForeignKey("file.id"), nullable=False, primary_key=True)
    archived_at = Column(types.DateTime(), nullable=True)
    retrieved_at = Column(types.DateTime(), nullable=True)

    file = orm.relationship("File", backref=backref("archive", uselist=False))


class Bundle(Model):
    """A general group of files."""

    __tablename__ = "bundle"

    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(64), unique=True, nullable=False)
    created_at = Column(types.DateTime, default=dt.datetime.now)
    versions = orm.relationship(
        "Version",
        backref="bundle",
        order_by="-Version.created_at",
        cascade="delete, save-update",
        cascade_backrefs=False,
    )


class Version(Model):
    """Keeps track of versions."""

    __tablename__ = "version"

    id = Column(types.Integer, primary_key=True)
    created_at = Column(types.DateTime, nullable=False)
    expires_at = Column(types.DateTime)
    included_at = Column(types.DateTime)
    removed_at = Column(types.DateTime)

    archived_at = Column(types.DateTime)
    archive_path = Column(types.String(256), unique=True)
    archive_checksum = Column(types.String(256), unique=True)

    bundle_id = Column(ForeignKey(Bundle.id, ondelete="CASCADE"), nullable=False)
    files = orm.relationship(
        "File", backref="version", cascade="delete, save-update", cascade_backrefs=False
    )

    app_root = None

    @property
    def relative_root_dir(self):
        """Build the relative root dir path for the bundle version."""
        return Path(self.bundle.name) / str(self.created_at.date())

    @property
    def full_path(self):
        """Returns the full path of the bundle"""
        return Path(self.app_root) / self.bundle.name / str(self.created_at.date())


class File(Model):
    """Represent a file."""

    __tablename__ = "file"

    id = Column(types.Integer, primary_key=True)
    path = Column(types.String(256), unique=True, nullable=False)
    checksum = Column(types.String(256))
    to_archive = Column(types.Boolean, nullable=False, default=False)

    version_id = Column(ForeignKey(Version.id, ondelete="CASCADE"), nullable=False)
    tags = orm.relationship(
        "Tag", secondary=file_tag_link, backref=backref("files", cascade_backrefs=False)
    )

    app_root = None

    @property
    def full_path(self) -> str:
        """Return the full path to the file."""
        if Path(self.path).is_absolute():
            return self.path
        return str(self.app_root / self.path)

    @property
    def is_included(self) -> bool:
        """Check if the file is included in Housekeeper."""
        return str(self.app_root) in self.full_path


class Tag(Model):
    """Represents tags for bundles."""

    __tablename__ = "tag"

    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(64), unique=True, nullable=False)
    category = Column(types.String(64))
    created_at = Column(types.DateTime, default=dt.datetime.now)



"""Module for deleting via CLI"""

import logging
import shutil
from pathlib import Path

import click

from housekeeper.date import get_date
from housekeeper.store.api.core import Store
from housekeeper.store.models import File, Tag

LOG = logging.getLogger(__name__)


@click.group()
def delete():
    """Delete things in the database."""


@delete.command("bundle")
@click.option("-y", "--yes", is_flag=True, help="skip checks")
@click.argument("bundle_name")
@click.pass_context
def bundle_cmd(context, yes, bundle_name):
    """Delete a empty bundle, that is a bundle without versions."""
    store: Store = context.obj["store"]
    bundle = store.get_bundle_by_name(bundle_name=bundle_name)
    if bundle is None:
        LOG.warning("bundle %s not found", bundle_name)
        raise click.Abort

    if bundle.versions:
        LOG.warning("Can not delete bundle, please remove all versions first")
        raise click.Abort

    question = f"Remove bundle {bundle.name} from database?"
    if not (yes or click.confirm(question)):
        raise click.Abort

    store.session.delete(bundle)
    store.session.commit()
    LOG.info(f"Deleted bundle {bundle.name}")


@delete.command("version")
@click.option("-b", "--bundle-name", help="Fetch all versions from a bundle")
@click.option("-i", "--version-id", type=int, help="Fetch a specific version")
@click.option("-y", "--yes", is_flag=True, help="skip checks")
@click.pass_context
def version_cmd(context, bundle_name, version_id, yes):
    """Delete a version from database."""
    store: Store = context.obj["store"]
    if not (bundle_name or version_id):
        LOG.info("Please select a bundle or a version")
        raise click.Abort

    if bundle_name:
        bundle = store.get_bundle_by_name(bundle_name=bundle_name)
        if not bundle:
            LOG.info(f"Could not find bundle {bundle_name}")
            return
        if len(bundle.versions) == 0:
            LOG.warning(f"Could not find versions for bundle {bundle_name}")
            return
        LOG.info(f"Deleting the latest version of bundle {bundle_name}")
        version_obj = bundle.versions[0]

    if version_id:
        version = store.get_version_by_id(version_id=version_id)
        if not version:
            LOG.warning(f"Could not find version {version_id}")
            raise click.Abort
        bundle = store.get_bundle_by_id(bundle_id=version.bundle_id)
        for ver in bundle.versions:
            if ver.id == version_id:
                version_obj = ver

    if version_obj.included_at:
        question = f"Remove bundle version {version_obj.full_path} from file system and database?"
    else:
        question = f"Remove bundle version {version_obj.created_at.date()} from database?"

    if not (yes or click.confirm(question)):
        raise click.Abort

    if version_obj.included_at:
        shutil.rmtree(version_obj.full_path, ignore_errors=True)

    store.session.delete(version_obj)
    store.session.commit()
    LOG.info(f"version deleted: {version_obj.full_path}")


@delete.command("files")
@click.option("--yes", is_flag=True, help="skip checks")
@click.option("-t", "--tag", multiple=True, help="file tag")
@click.option("-b", "--bundle-name", help="bundle name")
@click.option("-a", "--before", help="version created before...")
@click.option("-n", "--notondisk", is_flag=True, help="rm db entry from files not on disk")
@click.option("-l", "--list-files", is_flag=True, help="lists files that will be deleted")
@click.option("-L", "--list-files-verbose", is_flag=True, help="lists additional information")
@click.pass_context
def files_cmd(context, yes, tag, bundle_name, before, notondisk, list_files, list_files_verbose):
    """Delete files based on tags."""

    validate_delete_options(tag=tag, bundle_name=bundle_name)
    before_date = parse_date(before) if before else None
    store: Store = context.obj["store"]

    if bundle_name:
        validate_bundle_exists(store=store, bundle_name=bundle_name)

    files = store.get_files_before(bundle_name=bundle_name, tag_names=tag, before_date=before_date)

    if notondisk:
        files = store.get_files_not_on_disk(files=files)

    if not files:
        LOG.warning("No files found")
        raise click.Abort

    if list_files_verbose:
        list_files_verbosely(files=files)

    elif list_files:
        list_files_with_full_path(files=files)

    if not (yes or click.confirm(f"Are you sure you want to delete {len(files)} files?")):
        raise click.Abort

    for file in files:
        if yes or click.confirm(f"Remove file from disk and database: {file.full_path}?"):
            delete_file(file=file, store=store)


@delete.command("tag")
@click.option("-n", "--name", help="Name of the tag to delete")
@click.option("-y", "--yes", is_flag=True, help="skip checks")
@click.pass_context
def tag_cmd(context, yes, name: str):
    """Delete a tag from the database.
    Entries in the file_tag_link table associated with the tag will be removed."""
    store: Store = context.obj["store"]
    tag: Tag | None = store.get_tag(tag_name=name)
    if not tag:
        LOG.warning(f"Tag {name} not found")
        raise click.Abort

    tag_file_count: int = len(tag.files) if tag.files else 0
    question = f"Delete tag {name} with {tag_file_count} associated files?"

    if yes or click.confirm(question):
        store.session.delete(tag)
        store.session.commit()
        LOG.info(f"Tag {name} deleted")


def validate_delete_options(tag: str, bundle_name: str):
    """Validate delete options."""
    if not (tag or bundle_name):
        LOG.info("Please specify a bundle or a tag")
        raise click.Abort


def validate_bundle_exists(store: Store, bundle_name: str):
    """Validate bundle exists."""
    if not store.get_bundle_by_name(bundle_name=bundle_name):
        LOG.warning(f"Bundle {bundle_name} not found")
        raise click.Abort


def list_files_verbosely(files):
    """List files verbosely."""
    for file in files:
        tags = ", ".join(tag.name for tag in file.tags)
        click.echo(
            f"{click.style(str(file.id), fg='blue')} | {file.full_path} | "
            f"{click.style(tags, fg='yellow')}"
        )


def list_files_with_full_path(files):
    for file in files:
        click.echo(file.full_path)


def delete_file(file: File, store: Store):
    file_path = Path(file.full_path)
    if file_should_be_unlinked(file):
        file_path.unlink()

    store.session.delete(file)
    store.session.commit()
    LOG.info(f"{file.full_path} deleted")


def file_should_be_unlinked(file: File):
    """Check if file should be unlinked."""
    file_path = Path(file.full_path)
    return file.is_included and (file_path.exists() or file_path.is_symlink())


def parse_date(date: str):
    """Attempt to parse date in two different formats."""
    try:
        return get_date(date=date)
    except ValueError:
        return get_date(date=date, date_format="%Y-%m-%d %H:%M:%S")


@delete.command("file")
@click.option("-y", "--yes", is_flag=True, help="skip checks")
@click.argument("file_id", type=int)
@click.pass_context
def file_cmd(context, yes, file_id):
    """Delete a file."""
    store: Store = context.obj["store"]
    file = store.get_file_by_id(file_id=file_id)
    if not file:
        LOG.info("file not found")
        raise click.Abort

    if file.is_included:
        question = f"Remove file {file.full_path} from file system and database?"
    else:
        question = f"Remove file {file.full_path} from database?"

    if yes or click.confirm(question):
        if file.is_included and Path(file.full_path).exists():
            Path(file.full_path).unlink()

        store.session.delete(file)
        store.session.commit()
        LOG.info("file deleted")



"""Module for including files via CLI"""
import datetime as dt
import logging

import click
from housekeeper.constants import ROOT
from housekeeper.exc import VersionIncludedError
from housekeeper.include import include_version
from housekeeper.store.api.core import Store

LOG = logging.getLogger(__name__)


@click.command()
@click.option("--version-id", type=int, help="version id of the bundle version")
@click.argument("bundle_name", required=False)
@click.pass_context
def include(context: click.Context, bundle_name: str, version_id: int):
    """Include a bundle of files into the internal space.

    Use bundle name if you simply want to include the latest version.
    """
    LOG.info("Running include")
    store: Store = context.obj["store"]
    if not (version_id or bundle_name):
        LOG.warning("Please use bundle name or version-id")
        raise click.Abort

    if version_id:
        LOG.info("Use version %s", version_id)
        version_obj = store.get_version_by_id(version_id=version_id)
        if version_obj is None:
            LOG.warning("version not found")
            raise click.Abort

    if bundle_name:
        bundle = store.get_bundle_by_name(bundle_name=bundle_name)
        if bundle is None:
            LOG.warning("bundle %s not found", bundle_name)
            raise click.Abort

        if len(bundle.versions) == 0:
            LOG.error("Could not find any versions for bundle %s", bundle_name)
            raise click.Abort

        LOG.info("Including latest version for %s", bundle_name)
        version_obj = bundle.versions[0]

    try:
        include_version(context.obj[ROOT], version_obj)
    except VersionIncludedError as error:
        LOG.warning(error.message)
        raise click.Abort

    version_obj.included_at = dt.datetime.now()
    store.session.commit()
    click.echo(click.style("included all files!", fg="green"))



from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from housekeeper.store.models import Bundle


def filter_bundle_by_name(bundles: Query, bundle_name: str, **kwargs) -> Query:
    """Return bundle by bundle name."""
    return bundles.filter(Bundle.name == bundle_name)


def filter_bundle_by_id(bundles: Query, bundle_id: int, **kwargs) -> Query:
    """Return bundle by bundle id."""
    return bundles.filter(Bundle.id == bundle_id)


class BundleFilters(Enum):
    """Define Bundle filter functions."""

    FILTER_BY_NAME: Callable = filter_bundle_by_name
    FILTER_BY_ID: Callable = filter_bundle_by_id


def apply_bundle_filter(
    bundles: Query,
    filter_functions: list[Callable],
    bundle_name: str | None = None,
    bundle_id: int | None = None,
) -> Query:
    """Apply filtering functions and return filtered query."""
    for filter_function in filter_functions:
        bundles: Query = filter_function(
            bundles=bundles,
            bundle_name=bundle_name,
            bundle_id=bundle_id,
        )
    return bundles



from datetime import datetime
from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from housekeeper.store.models import Version


def filter_version_by_id(versions: Query, version_id: int, **kwargs) -> Query:
    """Return version by version id."""
    return versions.filter(Version.id == version_id)


def filter_version_by_date(versions: Query, before_date: datetime, **kwargs) -> Query:
    """Return version by date."""
    return versions.filter(Version.created_at < before_date)


class VersionFilter(Enum):
    """Define Version filter functions."""

    FILTER_BY_ID: Callable = filter_version_by_id
    FILTER_BY_DATE: Callable = filter_version_by_date


def apply_version_filter(
    versions: Query,
    filter_functions: list[Callable],
    version_id: int | None = None,
    before_date: datetime | None = None,
) -> Query:
    """Apply filtering functions and return filtered query."""
    for filter_function in filter_functions:
        versions: Query = filter_function(
            versions=versions, version_id=version_id, before_date=before_date
        )
    return versions
